@@ -2,6 +2,15 @@
 
 set -eux
 
+function copy_logs() {
+  log_dir=/host-rw/logs
+  mkdir -p ${log_dir}
+  cp -a ${data_dir}/log ${log_dir}/postgresql || :
+  mkdir -p ${log_dir}/pgroonga/ || :
+  cp -a ${data_dir}/pgroonga*.log* ${log_dir}/pgroonga/ || :
+  chmod -R go+rx ${log_dir}
+}
+
 function run_test() {
   echo "::group::Run test"
 
@@ -20,14 +29,28 @@ function run_test() {
     echo "::group::Diff"
     cat regression.diffs
     echo "::endgroup::"
-    mkdir -p /host-rw/logs
+    copy_logs
     cp -a regression.diffs /host-rw/logs/
-    cp -a ${data_dir}/log /host-rw/logs/postgresql || :
-    mkdir -p /host-rw/logs/pgroonga/ || :
-    cp -a ${data_dir}/pgroonga.log* /host-rw/logs/pgroonga/ || :
-    chmod -R go+rx /host-rw/logs/
     exit ${pg_regress_status}
   fi
+
+
+  echo "::group::Test primary-maintainer"
+
+  set +e
+  $(${pg_config} --bindir)/createuser postgres --superuser
+  /host/packages/test-primary-maintainer.sh ${pg_config}
+  test_status=$?
+  $(${pg_config} --bindir)/dropuser postgres
+  set -e
+
+  echo "::endgroup::"
+
+
+  if [ ${test_status} -ne 0 ]; then
+    copy_logs
+    exit ${test_status}
+  fi
 }
 
 echo "::group::Prepare repositories"
@@ -43,12 +66,11 @@ case ${os} in
         ;;
       *)
         DNF="dnf --enablerepo=crb"
-        ${DNF} install -y \
-               https://apache.jfrog.io/artifactory/arrow/${os}/${major_version}/apache-arrow-release-latest.rpm
         ;;
     esac
 
     ${DNF} install -y \
+           https://packages.apache.org/artifactory/arrow/${os}/${major_version}/apache-arrow-release-latest.rpm \
            https://download.postgresql.org/pub/repos/yum/reporpms/EL-${major_version}-x86_64/pgdg-redhat-repo-latest.noarch.rpm \
            https://packages.groonga.org/${os}/${major_version}/groonga-release-latest.noarch.rpm
     ;;
@@ -117,8 +139,10 @@ sudo -u postgres -H \
      --locale=C \
      --pgdata=${data_dir} \
      --username=root
-echo "max_prepared_transactions = 1" | \
-  sudo -u postgres -H tee --append ${data_dir}/postgresql.conf
+{
+  echo "max_prepared_transactions = 1"
+  echo "pgroonga.enable_wal = yes"
+} | sudo -u postgres -H tee --append ${data_dir}/postgresql.conf
 sudo -u postgres -H \
      $(${pg_config} --bindir)/pg_ctl start \
      --pgdata=${data_dir}
@@ -129,16 +153,6 @@ cp -a \
    /tmp/
 cd /tmp
 case "${os}" in
-  almalinux)
-    if [ ${postgresql_version} -lt 12 ]; then
-      rm sql/function/highlight-html/declarative-partitioning.sql
-      rm sql/function/wal-set-applied-position/declarative-partitioning.sql
-      rm sql/vacuum/two-phase-commit.sql
-    fi
-    if [ ${postgresql_version} -lt 13 ]; then
-      rm sql/full-text-search/text/single/declarative-partitioning.sql
-    fi
-    ;;
   fedora)
     # Require Groonga 10.1.0 or later.
     rm sql/function/highlight-html/one-keyword.sql
@@ -162,23 +176,46 @@ echo "::group::Upgrade"
 
 ${DNF} remove -y ${pgroonga_package}
 
-# Disable upgrade test for first time packages.
-# TODO: Automate this. We don't want to maintain this manually.
-# case ${postgresql_version} in
-#   16) # TODO: Remove this after 3.1.4 release.
-#     exit
-#     ;;
-# esac
+if ${DNF} info ${pgroonga_package} > /dev/null 2>&1; then
+  is_first_release=no
+else
+  is_first_release=yes
+fi
 
-${DNF} install -y ${pgroonga_package}
-createdb upgrade
-psql upgrade -c 'CREATE EXTENSION pgroonga'
-${DNF} install -y ${packages_dir}/*.rpm
-psql upgrade -c 'ALTER EXTENSION pgroonga UPDATE'
+if [ "${is_first_release}" = "yes" ]; then
+  echo "Skip because ${pgroonga_package} hasn't been released yet."
+else
+  ${DNF} install -y ${pgroonga_package}
+
+  # We can use "rpm -q" for the latest released ${pgroonga_package} while
+  # the latest released ${pgroonga_package} is installed.
+  #
+  # Example: postgresql16-pgdg-pgroonga-3.2.4-1.el8.x86_64
+  pgroonga_latest_released_package=$(rpm -q ${pgroonga_package})
+  # Example: 3.2.4
+  pgroonga_latest_released_version=$(rpm -q --queryformat="%{VERSION}" ${pgroonga_package})
+
+  createdb upgrade
+  psql upgrade -c 'CREATE EXTENSION pgroonga'
+  ${DNF} install -y ${packages_dir}/*.rpm
+  psql upgrade -c 'ALTER EXTENSION pgroonga UPDATE'
+fi
 
 echo "::endgroup::"
 
-run_test
+echo "::group::Downgrade"
+
+if [ "${is_first_release}" = "yes" ]; then
+  echo "Skip because ${pgroonga_package} hasn't been released yet."
+else
+  createdb downgrade
+  psql downgrade -c 'CREATE EXTENSION pgroonga'
+  psql downgrade \
+    -c "ALTER EXTENSION pgroonga UPDATE TO '${pgroonga_latest_released_version}'"
+  ${DNF} install -y ${pgroonga_latest_released_package}
+fi
+
+echo "::endgroup::"
 
 echo "::group::Postpare"
 
